import numpy as np
import pandas as pd
import os
import re
import xml.etree.cElementTree as ET
from io import StringIO

import utils
from cfatools.logreader.dataset import DatasetReader
from cfatools.logreader.pump_calib import PumpCalibReader
import cfatools.processor.iceblock as iceblock_processor
import cfatools.processor.pump as pump_processor
import cfatools.processor.encoder as encoder_processor


class IcbktransitionProvider:

    def __init__(self):
        self.dataset_reader = None
        self.iceblock_filename = ""
        self.iceblock_df = pd.DataFrame()
        self.transitions_df = pd.DataFrame()
        self.encoder_df = pd.DataFrame()

    def load_dataset(self, dataset_reader: DatasetReader):
        self.dataset_reader = dataset_reader
        self.iceblock_filename = os.path.join(self.dataset_reader.dataset_path, "processed", "iceblocks.csv")
        self.encoder_df = self.dataset_reader.get_timeseries("ENCODER_periodic")

    def build_iceblock_df(self, retrieve_iceblocks_from_processed_csv: bool = True) -> pd.DataFrame:
        if retrieve_iceblocks_from_processed_csv and os.path.exists(self.iceblock_filename):
            iceblock_df = pd.read_csv(self.iceblock_filename,
                                      parse_dates=["datetime_start", "datetime_end", "datetime_stacked"])
        else:
            iceblock_df = self.dataset_reader.get_timeseries("ICBKCTRL_instant")
            iceblock_df = iceblock_processor.get_clean_iceblock_df(iceblock_df)
        self.iceblock_df = iceblock_df
        return iceblock_df

    def save_iceblock_df_as_csv(self):
        self.iceblock_df.to_csv(self.iceblock_filename, index=False, date_format="%Y-%m-%dT%H:%M:%SZ")

    def get_transitions(self) -> pd.DataFrame:
        transitions_df = self.iceblock_df[["id", "datetime_start"]]
        transitions_df = transitions_df.rename(columns={"id": "to_id", "datetime_start": "datetime"})
        transitions_df["from_id"] = transitions_df["to_id"].shift()
        if self.iceblock_df.iloc[-1]["datetime_end"] is not pd.NaT:
            transitions_df = transitions_df.append(pd.DataFrame({"from_id": max(self.iceblock_df["id"]),
                                                                 "to_id": np.nan,
                                                                 "datetime": self.iceblock_df.iloc[-1]["datetime_end"]},
                                                                index=[len(transitions_df.index)]))
        transitions_df["timestamp"] = utils.pd_time_to_epoch_ms(transitions_df["datetime"])
        self.transitions_df = transitions_df
        return transitions_df

    def move_transition(self, transition_id: int, new_timestamp_sec: int):
        transition = self.transitions_df.iloc[transition_id]
        new_timestamp = pd.Timestamp(new_timestamp_sec * 10**9, tz="UTC")
        self.iceblock_df.loc[self.iceblock_df["id"] == transition["from_id"], "datetime_end"] = new_timestamp
        self.iceblock_df.loc[self.iceblock_df["id"] == transition["to_id"], "datetime_start"] = new_timestamp
        self.update_iceblocks_height()
        self.get_transitions()  # Updates self.transitions_df

    def delete_transition(self, transition_id: int):
        self.iceblock_df = self.iceblock_df.drop(transition_id)
        self.iceblock_df = self._rebuild_index__(self.iceblock_df)
        self.get_transitions()  # Updates self.transitions_df

    def _rebuild_index__(self, iceblock_df: pd.DataFrame) -> pd.DataFrame:
        iceblock_df.reset_index(drop=True, inplace=True)
        iceblock_df["id"] = range(1, len(iceblock_df.index) + 1)
        return iceblock_df

    def update_iceblocks_height(self):
        for index, iceblock in self.iceblock_df.iterrows():
            encoder_height = self.encoder_df.iloc[self.encoder_df.index.get_loc(iceblock["datetime_start"],
                                                                                method='nearest')]["position"]
            self.iceblock_df.loc[index, "initial_height"] = encoder_height

