import pandas as pd
import datetime
import os
import statsmodels.api as sm

from PyQt5.QtCore import QObject, pyqtSignal

from cfatools.logreader.dataset import DatasetReader


class PicarroProvider(QObject):

    sig_picarro_df_updated = pyqtSignal(pd.DataFrame, name="picarro_df_updated")
    sig_dataset_name_updated = pyqtSignal(str, name="dataset_name_updated")
    sig_calibs_df_updated = pyqtSignal(pd.DataFrame, name="calibs_df_updated")
    sig_near_days_range_updated = pyqtSignal(dict, name="near_days_range_updated")
    sig_near_calibs_df_updated = pyqtSignal(pd.DataFrame, name="near_calibs_df_updated")
    sig_stdtrue_df_updated = pyqtSignal(pd.DataFrame, name="stdtrue_df_updated")
    sig_near_calibs_df_with_truevalue_updated = pyqtSignal(pd.DataFrame, name="near_calibs_df_with_truevalue_updated")
    sig_equation_dict_updated = pyqtSignal(dict, name="equation_dict_updated")

    def __init__(self):
        super(PicarroProvider, self).__init__()

        self.dataset_path = ""

        # Picarro raw timeseries
        self.picarro_df = pd.DataFrame()

        # Standards' true value
        self.stdtrue_df = pd.DataFrame()

        # Picarro's calibrations' injections: all available for this Picarro
        self.calibs_df = pd.DataFrame()
        self.calibs_filename = ""

        # Picarro's calibrations' injections: only those close (time wise) to the current dataset
        self.near_calibs_df = pd.DataFrame()

        # Number of days before and after current dataset so a calib is considered as "near".
        self.near_days_range = {"before" : 5, "after": 5}

        # Correction's equation
        self.equation_dict = dict()

        self.sig_picarro_df_updated.connect(self.update_near_calibs_df)
        self.sig_calibs_df_updated.connect(self.update_near_calibs_df)
        self.sig_stdtrue_df_updated.connect(self.update_equation_dict)
        self.sig_near_calibs_df_updated.connect(self.add_truevalue_to_marked_calib_df)
        self.sig_stdtrue_df_updated.connect(self.add_truevalue_to_marked_calib_df)
        self.sig_near_calibs_df_with_truevalue_updated.connect(self.update_equation_dict)
        self.sig_near_days_range_updated.connect(self.update_near_calibs_df)

    def get_dataset_reader(self, dataset_path: str) -> DatasetReader:
        root_directory = os.path.dirname(dataset_path)
        dataset_name = os.path.basename(dataset_path)
        valid, error_msg = DatasetReader.dataset_name_is_valid(dataset_name)
        if not valid:
            raise ValueError(error_msg)
        dataset_reader = DatasetReader(base_path=root_directory, dataset=dataset_name)
        self.dataset_path = dataset_reader.dataset_path
        self.sig_dataset_name_updated.emit(dataset_name)
        return dataset_reader

    def update_picarro_df(self, dataset_path: str) -> pd.DataFrame:
        dataset_reader = self.get_dataset_reader(dataset_path)
        if "PICARRO_periodic" not in dataset_reader.get_instruments_names():
            raise ValueError("There is no Picarro log file in this dataset.")
        self.picarro_df = dataset_reader.get_timeseries("PICARRO_periodic")
        self.sig_picarro_df_updated.emit(self.picarro_df)
        return self.picarro_df

    def get_mean_and_sd(self, first_date: datetime.datetime, last_date: datetime.datetime) -> dict:
        output_dict = self.picarro_df[(self.picarro_df.index >= first_date)
                                      & (self.picarro_df.index <= last_date)].agg(['mean', 'std']).to_dict()
        return output_dict

    def get_df_from_dat(self, filenames: list) -> pd.DataFrame:
        """Convert Picarro's original .dat files into a single pd.DataFrame

        Parameters
        ----------
        filenames: list
            List of full file path to Picarro's `.dat` files.

        Returns
        -------
        pd.DataFrame
            DataFrame containing Picarro's raw data.
        """
        picarro_df = pd.DataFrame()
        for filename in filenames:
            step_df = pd.read_csv(filename, delim_whitespace=True)
            step_df = step_df[["DATE", "TIME", "H2O", "Delta_18_16", "Delta_D_H"]]
            picarro_df = picarro_df.append(step_df)

        picarro_df["datetime"] = pd.to_datetime(picarro_df["DATE"] + " " + picarro_df["TIME"]).dt.tz_localize('UTC')
        picarro_df = picarro_df.drop(columns=["DATE", "TIME"])

        picarro_df = picarro_df.sort_values(by=['datetime'])
        picarro_df = picarro_df.set_index("datetime")

        return picarro_df

    def update_calib_df(self, filename: str):
        if os.stat(filename).st_size == 0:
            return

        if os.path.exists(filename):
            self.calibs_filename = filename
            self.calibs_df = pd.read_csv(filename,
                                    sep="\t",
                                    parse_dates=["plateau_start", "plateau_end", "valid_start", "valid_end"])
            self.calibs_df = self.calibs_df.sort_values(by=['valid_start'])
            self.sig_calibs_df_updated.emit(self.calibs_df)
        else:
            raise FileNotFoundError(filename)

    def update_near_days_range(self, before_after: str, value: int):
        self.near_days_range[before_after] = value
        self.sig_near_days_range_updated.emit(self.near_days_range)

    def update_near_calibs_df(self):
        if (len(self.calibs_df.index) == 0) or (len(self.picarro_df.index) == 0):
            return

        min_date = self.picarro_df.index.min() - datetime.timedelta(days=self.near_days_range["before"])
        max_date = self.picarro_df.index.max() + datetime.timedelta(days=self.near_days_range["after"])
        near_calibs_df = self.calibs_df[(self.calibs_df["valid_start"] > min_date) &
                                        (self.calibs_df["valid_end"] < max_date)]

        self.near_calibs_df = near_calibs_df.reset_index().rename(columns={'index': 'original_index'})
        self.sig_near_calibs_df_updated.emit(self.near_calibs_df)

    def update_stdtrue_df(self, filename: str):
        self.stdtrue_df = pd.read_csv(filename, sep="\t")
        self.stdtrue_df["date"] = pd.to_datetime(self.stdtrue_df["date"] + "T00:00:00Z")
        self.sig_stdtrue_df_updated.emit(self.stdtrue_df)

    def mark_calib(self, index: int, keep: bool):
        original_index = self.near_calibs_df.at[index, 'original_index']
        self.calibs_df.at[original_index, "keep"] = keep

        # Update file
        if os.path.exists(self.calibs_filename):
            self.calibs_df.to_csv(self.calibs_filename,
                                  sep="\t",
                                  index=False)

        self.sig_calibs_df_updated.emit(self.calibs_df)

    def add_truevalue_to_marked_calib_df(self):
        if len(self.near_calibs_df.index) == 0:
            return

        for species in ["Delta_18_16", "Delta_D_H"]:
            if not (species + "_truevalue" in self.near_calibs_df.columns):
                self.near_calibs_df[species + "_truevalue"] = None
                self.near_calibs_df[species + "_truevalue_std"] = None

            for standard in self.near_calibs_df["standard"].unique():
                std_calibs_df = self.near_calibs_df[(self.near_calibs_df["standard"] == standard)
                                                    & self.near_calibs_df["keep"]]
                std_stdtrue_df = self.stdtrue_df[self.stdtrue_df["name"] == standard]
                for index, row in std_calibs_df.iterrows():
                    closest_stdtrue_df = std_stdtrue_df.iloc[
                        (std_stdtrue_df["date"] - row["valid_start"]).abs().argsort()[:1]]
                    truevalue_mean = closest_stdtrue_df[species + "_mean"].to_list()[0]
                    truevalue_std = closest_stdtrue_df[species + "_std"].to_list()[0]
                    self.near_calibs_df.loc[index, species + "_truevalue"] = truevalue_mean
                    self.near_calibs_df.loc[index, species + "_truevalue_std"] = truevalue_std

        self.sig_near_calibs_df_with_truevalue_updated.emit(self.near_calibs_df)

    def update_equation_dict(self):
        if len(self.near_calibs_df.index) < 2:
            return
        if not ("Delta_18_16_truevalue" in self.near_calibs_df.columns):
            return

        calibs_df = self.near_calibs_df[self.near_calibs_df["keep"]]

        for species in ["Delta_18_16", "Delta_D_H"]:
            meas = calibs_df[species + "_mean"].to_list()
            true = calibs_df[species + "_truevalue"].to_list()

            model = sm.OLS(true, sm.tools.add_constant(meas)).fit()

            self.equation_dict[species] = {"slope": model.params[1],
                                           "intercept": model.params[0]}

        self.sig_equation_dict_updated.emit(self.equation_dict)

    def save_calibration_points_and_equations(self) -> dict:
        save_dir = os.path.join(self.dataset_path, "processed")
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)

        # Save calibration points
        calibs_filepath = os.path.join(save_dir, "picarro_calib_points.txt")
        save_df = self.near_calibs_df[self.near_calibs_df["keep"]]
        save_df = save_df.drop(columns=["keep", "original_index"])
        save_df.to_csv(path_or_buf=calibs_filepath,
                       sep="\t",
                       index=False)

        # Save equations
        equations_filepath = os.path.join(save_dir, "picarro_calib_equations.txt")
        equation_df = pd.DataFrame.from_dict(self.equation_dict)
        equation_df.to_csv(path_or_buf=equations_filepath,
                           sep="\t",
                           index=True)

        filepaths_dict = {"calibs": calibs_filepath,
                         "equations": equations_filepath}

        return filepaths_dict
