import os
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

import utils
from config import Config
from dataprovider.picarroprovider import PicarroProvider
from gui.ui_mainwindow import Ui_MainWindow


class PicarroCalibUim(QObject):
    """Class managing the Widget which are common for all Picarro calib sub-tabs."""

    def __init__(self, picarro_prvd: PicarroProvider, main_ui: Ui_MainWindow, config: Config):
        super(PicarroCalibUim, self).__init__()
        self.main_ui = main_ui
        self.picarro_prvd = picarro_prvd
        self.config = config

        self.__init_calibrations_file__()
        self.__init_stdtrue_filename__()

        self.main_ui.picarro_pushbutton_dataset_load.clicked.connect(self.__load_dataset__)
        self.main_ui.picarro_pushbutton_stdtrue_select.clicked.connect(self.__select_stdtrue_file__)
        self.main_ui.picarro_pushbutton_calib_select.clicked.connect(self.__select_calibrations_file__)

        self.picarro_prvd.sig_dataset_name_updated.connect(self.main_ui.picarro_lineedit_dataset.setText)

    def __load_dataset__(self):
        # Get the directory containing the dataset's files.
        directory = utils.get_data_path(self.config, "datasets", "dataset")
        if directory == '':
            return

        try:
            self.picarro_prvd.update_picarro_df(directory)
        except ValueError as e:
            utils.show_popup(str(e), "Could not load data set!")
            return

    ####################################################################
    # Standard "true" value file management

    def __init_stdtrue_filename__(self):
        default_filename = os.path.join(self.config.read("DATA_SOURCE", "picarro_calibs_dir"),
                                        "standard_truevalue.txt")
        if os.path.exists(default_filename):
            self.read_stdtrue_file(default_filename)

    def __select_stdtrue_file__(self):
        # Get the directory containing all the data' directories of type `data_type`
        if os.path.exists(self.config.read("DATA_SOURCE", "picarro_calibs_dir")):
            base_dir = self.config.read("DATA_SOURCE", "picarro_calibs_dir")
        else:
            base_dir = os.path.expanduser("~")

        # Open a File Dialog to let the user choose the required data set.
        if "PYCHARM_HOSTED" in os.environ:
            filename = QFileDialog.getOpenFileName(
                caption="Select the file containing the Standard's `true` values",
                directory=base_dir,
                options=QFileDialog.DontUseNativeDialog,
            )[0]
        else:
            filename = QFileDialog.getOpenFileName(
                caption="Select the file containing the Standard's `true` values",
                directory=base_dir
            )[0]

        # Return if the user cancelled selection
        if filename == '':
            return ''

        # Update base dir value in config file, so the user
        self.config.write("DATA_SOURCE", "picarro_calibs_dir", os.path.dirname(filename))

        self.read_stdtrue_file(filename)

    def read_stdtrue_file(self, filename: str):
        self.main_ui.picarro_lineedit_stdtrue_filename.setText(filename)
        self.picarro_prvd.update_stdtrue_df(filename)

    ####################################################################
    # Calibration file management

    def __init_calibrations_file__(self):
        filename = self.config.read("DATA_SOURCE", "picarro_calib_file")
        if os.path.exists(filename):
            try:
                self.picarro_prvd.update_calib_df(filename)
            except Exception as e:
                utils.show_popup(str(e), "Could not load data set!")
            else:
                self.main_ui.picarro_lineedit_calib_filename.setText(filename)

    def __select_calibrations_file__(self):
        # Get the directory containing all the data' directories of type `data_type`
        if os.path.exists(self.config.read("DATA_SOURCE", "picarro_calib_file")):
            base_dir = os.path.dirname(self.config.read("DATA_SOURCE", "picarro_calib_file"))
        elif os.path.exists(self.config.read("DATA_SOURCE", "picarro_calibs_dir")):
            base_dir = self.config.read("DATA_SOURCE", "picarro_calibs_dir")
        else:
            base_dir = os.path.expanduser("~")

        # Open a File Dialog to let the user choose the required data set.
        if "PYCHARM_HOSTED" in os.environ:
            filename = QFileDialog.getOpenFileName(
                caption="Select the file containing the instrument's injections data",
                directory=base_dir,
                options=QFileDialog.DontUseNativeDialog,
            )[0]
        else:
            filename = QFileDialog.getOpenFileName(
                caption="Select the file containing the instrument's injections data",
                directory=base_dir
            )[0]

        # Return if the user cancelled selection
        if filename == '':
            return ''

        # Update base dir value in config file, so the user
        self.config.write("DATA_SOURCE", "picarro_calibs_dir", os.path.dirname(filename))
        self.config.write("DATA_SOURCE", "picarro_calib_file", filename)

        try:
            self.picarro_prvd.update_calib_df(filename)
        except Exception as e:
            utils.show_popup(str(e), "Could not load calibrtion file!")
        else:
            self.main_ui.picarro_lineedit_calib_filename.setText(filename)
