import pyqtgraph as pg
import pyqtgraph.exporters
import datetime
import pandas as pd
import numpy as np
import re
import os
from PyQt5.QtWidgets import QTableWidget, QFileDialog, QMessageBox
from PyQt5.QtGui import QColor

datetime_format = "%Y-%m-%dT%H:%M:%S.%fZ"


def pd_time_to_epoch_ms(series: pd.Series) -> list:
    """Convert a Pandas' Series containing Timestamp data into 1D array containing seconds (as float) since epoch."""
    # Pandas' Timestamp data are in nanoseconds.
    return list(pd.to_datetime(series).view(np.int64) / 1e9)


def datetime64_to_epoch_ms(datetimes_list: np.ndarray) -> list:
    return [float(instant) / 1e9 for instant in datetimes_list]


def epoch_to_string(epoch: float) -> str:
    return datetime.datetime.fromtimestamp(epoch, tz=datetime.timezone.utc).strftime(datetime_format)


def convert_to_numeric(str_value: str):
    """Convert a string to float (if there is a "." in the string) or to integer, if possible."""
    if re.match("^-?\d+\.\d+$", str_value):
        return float(str_value)
    elif re.match("^-?\d+$", str_value):
        return int(str_value)
    else:
        return str_value


####################################################################
# Graphics

class TimeAxisItem(pg.AxisItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.setLabel(text='Time', units=None)
        self.enableAutoSIPrefix(False)

    def tickStrings(self, values, scale, spacing):
        try:
            tick_strings = [datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc).strftime("%H:%M:%S") for value in values]
        except ValueError:
            tick_strings = ["ERROR" for value in values]
        return tick_strings


def clear_legend_items(legend_item: pg.LegendItem) -> pg.LegendItem:
    """Remove all the elements of the input `legend_item`.

    Parameters
    ----------
    legend_item: pg.LegendItem
        Legend item to be emptied.

    Returns
    -------
    pg.LegendItem:
        Same as input, but empty.
    """
    for sample, label in legend_item.items:
        legend_item.removeItem(label.text)
    return legend_item


def export_plot(plot_item: pg.PlotItem, graphics_view: pg.PlotWidget, base_dir: str) -> None:
    """Export the `plot_item` as a PNG file.

    This is a convenience function for exporting plots with pre-set parameters. The same thing can be achieved by
    right-clicking the plot and using build-in pyqtgraph's contextual menu. Except that the axis' color cannot be
    modified with the right-click button.

    Parameters
    ----------
    plot_item: pg.PlotItem
        Plot item to be exported as PNG file.
    graphics_view: pg.PlotWidget
        Widget where the `plot_item` is. Used to determine plot's dimension in the GUI.

    Returns
    -------
    None
        Creates a PNG file. A file dialog asks the user the output file name/path.
    """
    # Create the default destination folder if it does not already exists.
    if not os.path.exists(base_dir):
        os.mkdir(base_dir)

    # Popup a FileDialog so the user can choose where to save the exported plot.
    filename, ext_filter = QFileDialog.getSaveFileName(caption="Export plot as PNG", filter="*.png", directory=base_dir)

    # Plot export cancelled by the user.
    if filename == "":
        return

    # Add png file extension if missing.
    file_extension = os.path.splitext(filename)
    if file_extension[1] != ".png":
        filename += ".png"

    exporter = pg.exporters.ImageExporter(plot_item)

    exporter.params.param('width').setValue(graphics_view.frameGeometry().width(),
                                            blockSignal=exporter.widthChanged)
    exporter.params.param('height').setValue(graphics_view.frameGeometry().height(),
                                             blockSignal=exporter.heightChanged)
    exporter.params.param('background').setValue(QColor("white"))

    # Change Axis' pen to have it black (more readable on white background than the original grey)
    gui_pen = plot_item.getAxis('left').pen()
    png_pen = pg.mkPen(color=(0, 0, 0))
    plot_item.getAxis('left').setPen(png_pen)
    plot_item.getAxis('bottom').setPen(png_pen)

    # Save to file
    # /!\ Note: there is a bug in pyqtgraph version 0.10.0, which lead to the "TypeError: 'float' object cannot be
    # interpreted as an integer" error.
    # Line 70 of file pyqtgraph.exporters.ImageExporter should be modified as follows:
    # bg = np.empty((int(self.params['width']), int(self.params['height']), 4), dtype=np.ubyte)
    exporter.export(filename)

    # Reset Axis' pen to previous pen
    plot_item.getAxis('left').setPen(gui_pen)
    plot_item.getAxis('bottom').setPen(gui_pen)


def highlight_row(table: QTableWidget, highlighted_row_id: int,
                  highlighted_row_color: QColor = QColor(173, 216, 230),
                  other_rows_color: QColor = QColor(255, 255, 255)) -> None:
    """Highlight a table's row by changing its background color.

    Parameters
    ----------
    table: PyQt5.QtWidgets.QTableWidget
        Table whose row should be highlighted.
    highlighted_row_id: int
        Index (first row is 0) of the row which should be highlighted.
    highlighted_row_color: PyQt5.QtGui.QColor
        Background color of the highlighted row. Default is light blue.
    other_rows_color: PyQt5.QtGui.QColor
        Background color of the other rows of the table. Default is white.

    Returns
    -------
        Nothing, modify the input ``table``
    """
    for row_id in range(table.rowCount()):
        if row_id == highlighted_row_id:
            row_color = highlighted_row_color
        else:
            row_color = other_rows_color
        for column_id in range(table.columnCount()):
            table.item(row_id, column_id).setBackground(row_color)


def get_data_path(config, data_type: str, containing: str) -> str:
    # Get the directory containing all the data' directories of type `data_type`
    if os.path.exists(config.read("DATA_SOURCE", data_type + "_dir")):
        base_dir = config.read("DATA_SOURCE", data_type + "_dir")
    else:
        base_dir = os.path.expanduser("~")

    # Open a File Dialog to let the user choose the required data set.
    dataset_path = QFileDialog.getExistingDirectory(
        caption="Select the folder containing " + containing + " files",
        directory=base_dir,
        options=QFileDialog.ShowDirsOnly
    )

    # Return if the user cancelled selection
    if dataset_path == '':
        return ''

    # Update base dir value in config file, so the user
    config.write("DATA_SOURCE", data_type + "_dir", os.path.dirname(dataset_path))

    return dataset_path


def show_popup(text: str, title: str = "Error"):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Warning)
    msg.setWindowTitle(title)
    msg.setText(str(text))
    msg.exec_()
